#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "pac.h"

/*! Error message */
static char error_msg[256] = { 0 };

/*! Tables initialised? */
static char has_init = 0;

/*! Table with Rice codes */
static PacRice_t rice_codes[16][256];

/*! Bits needed to encode a certain range */
static char bits_range[256];

static char inline
read_bit (PacObject_t *obj);

static int inline
read_bits (PacObject_t *obj, int count);

static void inline
write_bit (PacObject_t *obj, char bit);

static void inline
write_bits (PacObject_t *obj, int bits, int count);

static void
flush_bits (PacObject_t *obj);

static void
create_rice_codes_table ();

static void
create_bits_range_table ();

static void
create_rice_k_table (PacObject_t *obj);

static void
update_rice_stats (PacObject_t *obj, int p, int n1, int n2);

static void
encode_pixel (PacObject_t *obj, int p, int n1, int n2);

static unsigned char
decode_pixel (PacObject_t *obj, int n1, int n2);


PacObject_t *
pac_new_obj ()
{
  PacObject_t *obj;
  
  obj = calloc (sizeof(PacObject_t), 1);
  if (! obj)
    {
      pac_set_error ("Out of memory!");
      return NULL;
    }
  
  if (! has_init)
    {
      create_rice_codes_table ();
      create_bits_range_table ();
      has_init = 1;
    }
  
  return obj;
}


void
pac_delete_obj (PacObject_t *obj)
{
  free (obj);
}


void
pac_reset_obj (PacObject_t *obj)
{
  memset (obj, 0, sizeof(PacObject_t));
}


void
pac_progress_callback (PacObject_t *obj, void (*callback)(double))
{
  obj->callback = callback;
}


int
pac_encode_to_file (PacObject_t *obj, PacImage_t *image, const char *dst)
{
  PacHeader_t header = { { 'P', 'A', 'C' }, '0' };
  unsigned char *c;
  int x, y, n1_off, n2_off, count;
  int components = image->format + 1;
  int progress = 1, max_progress = image->height * 2;
  
  /* Open output file */
  if (! (obj->f = fopen (dst, "wb")))
    {
      pac_set_error ("Could not open %s", dst);
      return 1;
    }
  
  /* Performance thingy */
  obj->max_k_value = image->bits_per_component;
  if (obj->max_k_value > 15)
    obj->max_k_value = 15;
  
  /* Create header */
  header.format = image->format;
  header.bits_per_component = image->bits_per_component;
  header.width_h = image->width >> 8;
  header.width_l = image->width & 0xFF;
  header.height_h = image->height >> 8;
  header.height_l = image->height & 0xFF;
  
  /* Write header to file */
  fwrite (&header, 1, sizeof(PacHeader_t), obj->f);
  
  /* Collect rice statistics */
  count = (image->width - 2) * components;
  n1_off = -components;
  n2_off = -components * 2;
  c = image->data + 2 * components;
  for (; count > 0; count--)
    {
      update_rice_stats (obj, *c , c[n1_off], c[n2_off]);
      c++;
    }
  
  for (y = 1; y < image->height; y++)
    {
      c = image->data + y * image->bytes_per_row;
      
      n1_off = -image->bytes_per_row;
      n2_off = (image->width == 1) ? n1_off : n1_off + components;
      for (count = components; count > 0; count--)
	{
	  update_rice_stats (obj, *c, c[n1_off], c[n2_off]);
	  c++;
	}
      
      count = (image->width - 1) * components;
      n1_off = -components;
      n2_off = -image->bytes_per_row;
      for (; count > 0; count--)
	{
	  update_rice_stats (obj, *c, c[n1_off], c[n2_off]);
	  c++;
	}
      
      /* Update progress bar every 16th row */
      if ((! (progress & 0xF)) && obj->callback)
	obj->callback ((double)progress / (double)max_progress);
      
      progress++;
    }
  
  /* Create rice k table */
  create_rice_k_table (obj);
  
  /* Write rice k table to file */
  for (x = 0; x < 256; x++)
    write_bits (obj, obj->rice_k[x], 4);
  
  /* Encode pixels */
  count = (image->width < 2) ? image->width : 2;
  count *= components;
  c = image->data;
  for (; count > 0; count--)
    {
      write_bits (obj, *c, 8);
      c++;
    }

  count = (image->width - 2) * components;
  n1_off = -components;
  n2_off = -components * 2;
  c = image->data + 2 * components;
  for (; count > 0; count--)
    {
      encode_pixel (obj, *c , c[n1_off], c[n2_off]);
      c++;
    }
  
  progress++;
  
  for (y = 1; y < image->height; y++)
    {
      c = image->data + y * image->bytes_per_row;
      
      n1_off = -image->bytes_per_row;
      n2_off = (image->width == 1) ? n1_off : n1_off + components;
      for (count = components; count > 0; count--)
	{
	  encode_pixel (obj, *c, c[n1_off], c[n2_off]);
	  c++;
	}
      
      count = (image->width - 1) * components;
      n1_off = -components;
      n2_off = -image->bytes_per_row;
      for (; count > 0; count--)
	{
	  encode_pixel (obj, *c, c[n1_off], c[n2_off]);
	  c++;
	}
      
      /* Update progress bar every 16th row */
      if ((! (progress & 0xF)) && obj->callback)
	obj->callback ((double)progress / (double)max_progress);
      
      progress++;
    }
  
  if (obj->callback)
    obj->callback (1.0);
  
  flush_bits (obj);
  fclose (obj->f);
  
  return 0;
}


PacImage_t *
pac_decode_from_file (PacObject_t *obj, const char *src)
{
  PacImage_t *image;
  PacHeader_t header;
  unsigned char *c;
  int count, components, width, height, progress = 1;
  int i, y, n1_off, n2_off;
  
  /* Open input file */
  if (! (obj->f = fopen (src, "rb")))
    {
      pac_set_error ("Could not open %s", src);
      return NULL;
    }
  
  if (fread (&header, 1, sizeof(PacHeader_t), obj->f) != sizeof(PacHeader_t))
    {
      pac_set_error ("Could not read PAC header from %s", src);
      fclose (obj->f);
      return NULL;
    }
  
  if (memcmp (&header, "PAC0", 4))
    {
      pac_set_error ("Invalid PAC magic in header from %s", src);
      fclose (obj->f);
      return NULL;
    }
  
  if (header.format < 0 || header.format > 3 ||
      header.bits_per_component < 8 || header.bits_per_component > 16)
    {
      pac_set_error ("Invalid PAC header data from %s", src);
      fclose (obj->f);
      return NULL;
    }
  
  width = (int)header.width_h << 8 | header.width_l;
  height = (int)header.height_h << 8 | header.height_l;
  
  image = pac_new_image (header.format, header.bits_per_component,
			 width, height);
  if (! image)
    {
      fclose (obj->f);
      return NULL;
    }
  
  /* Read Rice k table from image (always 256 entries) */
  for (i = 0; i < 256; i++)
    obj->rice_k[i] = read_bits (obj, 4);
  
  /* Decode pixels */
  components = image->format + 1;
  
  count = (image->width < 2) ? image->width : 2;
  count *= components;
  c = image->data;
  for (; count > 0; count--)
    *c++ = read_bits (obj, 8);
  
  count = (image->width - 2) * components;
  n1_off = -components;
  n2_off = -components * 2;
  c = image->data + 2 * components;
  for (; count > 0; count--)
    {
      *c = decode_pixel (obj, c[n1_off], c[n2_off]);
      c++;
    }
  
  for (y = 1; y < image->height; y++)
    {
      c = image->data + y * image->bytes_per_row;
      
      n1_off = -image->bytes_per_row;
      n2_off = (image->width == 1) ? n1_off : n1_off + components;
      for (count = components; count > 0; count--)
	{
	  *c = decode_pixel (obj, c[n1_off], c[n2_off]);
	  c++;
	}
      
      count = (image->width - 1) * components;
      n1_off = -components;
      n2_off = -image->bytes_per_row;
      for (; count > 0; count--)
	{
	  *c = decode_pixel (obj, c[n1_off], c[n2_off]);
	  c++;
	}
      
      /* Update progress bar every 16th row */
      if ((! (progress & 0xF)) && obj->callback)
	obj->callback ((double)progress / (double)image->height);
      
      progress++;
    }
  
  if (obj->callback)
    obj->callback (1.0);
  
  fclose (obj->f);
  
  return image;
}


PacImage_t *
pac_new_image (PacFormat_t format, int bits_per_component,
	       int width, int height)
{
  int bytes;
  PacImage_t *image;
  
  image = malloc (sizeof(PacImage_t));
  if (! image)
    {
      pac_set_error ("Out of memory!");
      return NULL;
    }
  
  /* Round to nearest byte */
  bytes = (bits_per_component + 7) >> 3;
  
  image->data = malloc (width * height * bytes * (format + 1));
  if (! image->data)
    {
      pac_set_error ("Out of memory!");
      free (image);
      return NULL;
    }
  
  image->format = format;
  image->bits_per_component = bits_per_component;
  image->bytes_per_pixel = bytes * (format + 1);
  image->bytes_per_row = bytes * (format + 1) * width;
  image->width = width;
  image->height = height;
  
  return image;
}


void
pac_delete_image (PacImage_t *image)
{
  free (image->data);
  free (image);
}


void
pac_print_error ()
{
  fprintf (stderr, "E: %s\n", error_msg);
}


char *
pac_get_error ()
{
  return error_msg;
}


/* PRIVATE */

void
pac_set_error (const char *fmt, ...)
{
  va_list ap;
  va_start (ap, fmt);
  vsnprintf (error_msg, 255, fmt, ap);
  va_end (ap);
}


/*! Read one bit from bit bucket */
static char inline
read_bit (PacObject_t *obj)
{
  PacBitBucket_t *bucket = &obj->bucket;
  int pos = bucket->pos;
  unsigned char bit, *cell;
  
  /* Refill bucket, no check for EOF because we need max
     performance here */
  if (pos == (PAC_BIT_BUCKET_SIZE * 8))
    pos = 0;
  
  if (! pos)
    fread (bucket->mem, 1, PAC_BIT_BUCKET_SIZE, obj->f);
  
  bucket->pos = pos + 1;
  cell = bucket->mem + (pos >> 3);
  bit = *cell & 0x80;
  *cell <<= 1;
  
  return bit;
}


/*! Read bits from bit bucket */
static int inline
read_bits (PacObject_t *obj, int count)
{
  PacBitBucket_t *bucket = &obj->bucket;
  int pos = bucket->pos, bits = 0, remaining_bits;
  unsigned char *cell;
  
  while (count)
    {
      /* Refill bucket, no check for EOF because we need max
	 performance here */
      if (pos == (PAC_BIT_BUCKET_SIZE * 8))
	pos = 0;
      
      if (! pos)
	fread (bucket->mem, 1, PAC_BIT_BUCKET_SIZE, obj->f);
      
      remaining_bits = 8 - (pos & 7);
      if (remaining_bits > count)
	remaining_bits = count;
      
      cell = bucket->mem + (pos >> 3);
      bits = (bits | *cell) << remaining_bits;
      *cell <<= remaining_bits;
      pos += remaining_bits;
      count -= remaining_bits;
    }
  
  bucket->pos = pos;
  
  return (bits >> 8);
}


/*! Write one bit to bit bucket */
static void inline
write_bit (PacObject_t *obj, char bit)
{
  PacBitBucket_t *bucket = &obj->bucket;
  int pos = bucket->pos;
  unsigned char *cell;
  
  /* Refill bucket, no check for EOF because we need max
     performance here */
  if (pos == (PAC_BIT_BUCKET_SIZE * 8))
    {
      fwrite (bucket->mem, 1, PAC_BIT_BUCKET_SIZE, obj->f);
      pos = 0;
    }
  
  cell = bucket->mem + (pos >> 3);
  *cell = (*cell << 1) | bit;
  bucket->pos = pos + 1;
}


/*! Write bits to bit bucket */
static void inline
write_bits (PacObject_t *obj, int bits, int count)
{
  PacBitBucket_t *bucket = &obj->bucket;
  static int mask[9] = { 0x00, 0x01, 0x03, 0x07, 0x0F,
			 0x1F, 0x3F, 0x7F, 0xFF };
  int pos = bucket->pos, remaining_bits, shifted;
  unsigned char *cell;
  
  while (count)
    {
      /* Refill bucket, no check for EOF because we need max
	 performance here */
      if (pos == (PAC_BIT_BUCKET_SIZE * 8))
	{
	  fwrite (bucket->mem, 1, PAC_BIT_BUCKET_SIZE, obj->f);
	  pos = 0;
	}
      
      remaining_bits = 8 - (pos & 7);
      if (remaining_bits > count)
	remaining_bits = count;
      
      cell = bucket->mem + (pos >> 3);
      shifted = bits >> (count - remaining_bits);
      *cell = (*cell << remaining_bits) | (shifted & mask[remaining_bits]);
      pos += remaining_bits;
      count -= remaining_bits;
    }
  
  bucket->pos = pos;
}


/*! Flush remaining bits in bucket to disk */
static void
flush_bits (PacObject_t *obj)
{
  PacBitBucket_t *bucket = &obj->bucket;
  unsigned char *cell = bucket->mem + (bucket->pos >> 3);
  int remaining_bits;
  
  /* Shift stray bits in place */
  remaining_bits = 8 - (bucket->pos & 7);
  *cell <<= remaining_bits;
  
  if (bucket->pos)
    fwrite (bucket->mem, 1, (bucket->pos + 7) >> 3, obj->f);
}


static void
create_rice_codes_table ()
{
  int x, i, q, k, len;
  int mask[16];
  
  x = 0;
  for (i = 0; i < 16; i++)
    {
      mask[i] = x;
      x = (x << 1) | 1;
    }
  
  for (k = 0; k < 16; k++)
    {
      for (x = 0; x < 256; x++)
	{
	  q = x >> k;
	  
	  len = q + 1 + k;
	  rice_codes[k][x].len = len;
	  rice_codes[k][x].code = (-2 << k) | (x & mask[k]);
	}
    }
}


static void
create_bits_range_table ()
{
  int pos = 0, max, bits;
  
  for (bits = 1; bits <= 8; bits++)
    {
      max = 1 << bits;
      
      for (; pos < max; pos++)
	bits_range[pos] = bits;
    }
}


static void
create_rice_k_table (PacObject_t *obj)
{
  int i, k, k_min, max_offset;
  
  for (max_offset = 0; max_offset < 256; max_offset++)
    {
      k = 0;
      k_min = obj->rice_stats[k][max_offset];
      for (i = 1; i <= obj->max_k_value; i++)
	{
	  if (obj->rice_stats[i][max_offset] < k_min)
	    {
	      k_min = obj->rice_stats[i][max_offset];
	      k = i;
	    }
	}
      
      obj->rice_k[max_offset] = k;
    }
}


static void
update_rice_stats (PacObject_t *obj, int p, int n1, int n2)
{
  int k, min, max, offset, max_offset;
  
  min = n1 < n2 ? n1 : n2;
  max = n1 > n2 ? n1 : n2;
  
  if (p < min || p > max)
    {
      if (p < min)
	{
	  max_offset = min - 1;
	  offset = min - p - 1;
	}
      else
	{
	  max_offset = 254 - max;
	  offset = p - max - 1;
	}
      
      for (k = 0; k <= obj->max_k_value; k++)
	obj->rice_stats[k][max_offset] += rice_codes[k][offset].len;
    }
}


static void
encode_pixel (PacObject_t *obj, int p, int n1, int n2)
{
  int k, min, max, max_offset, offset, delta, len, leading_ones;
  
  min = n1 < n2 ? n1 : n2;
  max = n1 > n2 ? n1 : n2;
  delta = max - min;
  
  if (p < min || p > max)
    {
      /* Out of range: 1 */
      write_bit (obj, 1);
      
      if (p < min)
	{
	  /* Below: 0 */
	  write_bit (obj, 0);
	  max_offset = min - 1;
	  offset = min - p - 1;
	}
      else
	{
	  /* Above: 1 */
	  write_bit (obj, 1);
	  max_offset = 254 - max;
	  offset = p - max - 1;
	}
      
      k = obj->rice_k[max_offset];
      len = rice_codes[k][offset].len;
      
      /* Only last 32 bits are stored in the Rice table.
	 Add leading ones if necessary. */
      leading_ones = len - 32;
      while (leading_ones > 0)
	{
	  if (leading_ones > 32)
	    {
	      write_bits (obj, -1, 32);
	      leading_ones -= 32;
	    }
	  else
	    {
	      write_bits (obj, -1, leading_ones);
	      leading_ones = 0;
	    }
	}
      
      if (len > 32)
	len = 32;
      
      write_bits (obj, rice_codes[k][offset].code, len);
    }
  else
    {
      /* In range: 0 */
      write_bit (obj, 0);
      
      /* If min == max we don't have to encode anything but the range bit */
      if (min != max)
	write_bits (obj, p - min, bits_range[delta]);
    }
}


static unsigned char
decode_pixel (PacObject_t *obj, int n1, int n2)
{
  int min, max, delta, offset, k, leading_ones;
  unsigned char p;
  
  min = n1 < n2 ? n1 : n2;
  max = n1 > n2 ? n1 : n2;
  delta = max - min;
  
  if (read_bit (obj))
    {
      /* Out of range */
      if (read_bit (obj))
	{
	  /* Above */
	  k = obj->rice_k[254 - max];
	  leading_ones = 0;
	  while (read_bit (obj))
	    leading_ones++;
	  
	  if (k)
	    offset = (leading_ones << k) + read_bits (obj, k);
	  else
	    offset = leading_ones;
	  
	  p = max + offset + 1;
	}
      else
	{
	  /* Below */
	  k = obj->rice_k[min - 1];
	  leading_ones = 0;
	  while (read_bit (obj))
	    leading_ones++;
	  
	  if (k)
	    offset = (leading_ones << k) + read_bits (obj, k);
	  else
	    offset = leading_ones;
	  
	  p = min - offset - 1;
	}
    }
  else
    {
      /* In range */
      if (delta)
	p = min + read_bits (obj, bits_range[delta]);
      else
	p = min;
    }
  
  return p;
}
