#ifndef _PAC_H
#define _PAC_H

/*! Size of bit bucket size in bytes */
#define PAC_BIT_BUCKET_SIZE  4096


/*! Pixel formats */
typedef enum
{
  PAC_GRAY,
  PAC_GRAY_ALPHA,
  PAC_RGB,
  PAC_RGB_ALPHA
}
PacFormat_t;


/*! Image feed to the encoder or as produced by the decoder */
typedef struct
{
  PacFormat_t format;
  int bits_per_component;
  int bytes_per_pixel;    /*!< Convenience variable */
  int bytes_per_row;      /*!< Convenience variable */
  int width;
  int height;
  unsigned char *data;
}
PacImage_t;


/*! Rice code entry */
typedef struct
{
  int len;   /*!< Code length */
  int code;  /*!< First 32 bits of code */
}
PacRice_t;


/*! Storage for coded data */
typedef struct
{
  char mem[PAC_BIT_BUCKET_SIZE];
  int pos;
}
PacBitBucket_t;


/*! Data container for PAC routines */
typedef struct
{
  /* Modified by encoder and decoder */
  PacBitBucket_t bucket;      /*!< Storage for coded data */
  int rice_stats[16][256];    /*!< Statistics for best k values */
  char rice_k[256];           /*!< Best k values for given offset */
  
  FILE *f;                    /*!< File descriptor */
  void (*callback)(double);   /*!< Progress callback handler */
  int max_k_value;            /*!< Max rice k value */
  int bits_per_component;     /*!< Same as in image */
}
PacObject_t;


/*! Header for PAC file */
typedef struct
{
  char magic[3];            /*!< The magic is the three characters PAC */
  char version;             /*!< Only version '0' exists so far */
  char format;              /*!< Pixel format */
  char bits_per_component;  /*!< Should be a value between 8 and 16 */
  unsigned char width_h;    /*!< Upper byte of width */
  unsigned char width_l;    /*!< Lower byte of width */
  unsigned char height_h;   /*!< Upper byte of height */
  unsigned char height_l;   /*!< Lower byte of height */
}
PacHeader_t;


/* PUBLIC */

/*! Create data container for PAC routines */
PacObject_t *
pac_new_obj ();

/*! Destroy data container */
void
pac_delete_obj (PacObject_t *obj);

/*! Reset container if you wish ro reuse it */
void
pac_reset_obj (PacObject_t *obj);

/*! Register progress handler */
void
pac_progress_callback (PacObject_t *obj, void (*callback)(double));

/*! Encode image to file */
int
pac_encode_to_file (PacObject_t *obj, PacImage_t *image, const char *dst);

/*! Decode image from file */
PacImage_t *
pac_decode_from_file (PacObject_t *obj, const char *src);

/*! Create image */
PacImage_t *
pac_new_image (PacFormat_t format, int bits_per_component,
	       int width, int height);

/*! Delete image */
void
pac_delete_image (PacImage_t *image);

/*! Print last error */
void
pac_print_error ();

/*! Return string to last error */
char *
pac_get_error ();


/* PRIVATE */

/*! Set error message, msg should be a pointer to static data */
void
pac_set_error (const char *fmt, ...);

/*! Load PNG image (png.c) */
PacImage_t *
png_decode_from_file (const char *src);

/*! Save PNG image (png.c) */
int
png_encode_to_file (PacImage_t *image, const char *dst);

#endif /* _PAC_H */
