/* PAC image loader plug-in for Gimp 1.2 and Gimp 2.0.
 *
 * Compile with -DPAC_GIMP_12 for Gimp 1.2 plug-in.
 *              -DPAC_GIMP_20 for Gimp 2.0 plug-in.
 *
 * Based on the BMP plug-in loader by Alexander Schulz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#ifdef PAC_GIMP_12
#include <libgimp/gimpintl.h>
#else
#include <libintl.h>
#define _(String) gettext(String)
#endif

#include "pac.h"

#define PLUG_IN_VERSION  "0.1 - 28 April 2004"

#ifdef PAC_GIMP_12
/* From "libgimp/stdplugins-intl.h" */

#define INIT_I18N()	G_STMT_START{			\
  setlocale(LC_MESSAGES, ""); 				\
  bindtextdomain("gimp-libgimp", LOCALEDIR);            \
  bindtextdomain("gimp-std-plugins", LOCALEDIR);	\
  textdomain("gimp-std-plugins");			\
  			}G_STMT_END

#define INIT_I18N_UI()	G_STMT_START{	\
  gtk_set_locale();			\
  setlocale (LC_NUMERIC, "C");		\
  INIT_I18N();				\
			}G_STMT_END
#endif

/* Local functions. */

static void   query      (void);

static void   run        (const gchar   *name,
                          gint     nparams,
                          const GimpParam  *param,
                          gint    *nreturn_vals,
                          GimpParam **return_vals);

static gint32 
read_pac (gchar *filename, int interactive);

static GimpPDBStatusType
write_pac (gchar  *filename,
	   gint32  gimp_image,
	   gint32  drawable_ID,
	   int interactive);


GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init_proc  */
  NULL,  /* quit_proc  */
  query, /* query_proc */
  run,   /* run_proc   */
};


MAIN ()


static void
query (void)
{
  static GimpParamDef load_args[] =
  {
    { GIMP_PDB_INT32,    "run_mode",     "Interactive, non-interactive" },
    { GIMP_PDB_STRING,   "filename",     "The name of the file to load" },
    { GIMP_PDB_STRING,   "raw_filename", "The name entered" },
  };
  static GimpParamDef load_return_vals[] =
  {
    { GIMP_PDB_IMAGE, "image", "Output image" },
  };
  static gint nload_args = sizeof (load_args) / sizeof (load_args[0]);
  static gint nload_return_vals = (sizeof (load_return_vals) /
				   sizeof (load_return_vals[0]));

  static GimpParamDef save_args[] =
  {
    { GIMP_PDB_INT32,    "run_mode",     "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE,    "image",        "Input image" },
    { GIMP_PDB_DRAWABLE, "drawable",     "Drawable to save" },
    { GIMP_PDB_STRING,   "filename",     "The name of the file to save the image in" },
    { GIMP_PDB_STRING,   "raw_filename", "The name entered" },
  };
  static gint nsave_args = sizeof (save_args) / sizeof (save_args[0]);

#ifdef PAC_GIMP_12
  INIT_I18N();
#endif

  gimp_install_procedure ("file_pac_load",
                          "Loads files of PAC file format",
                          "This plug-in loads PAC graphic format files.",
                          "Johan Bolmsjo <johan@nocrew.org>",
                          "Copyright 2002 by Johan Bolmsjo",
                          PLUG_IN_VERSION,
                          "<Load>/PAC",
                          NULL,
                          GIMP_PLUGIN,
                          nload_args,
			  nload_return_vals,
                          load_args,
			  load_return_vals);

  gimp_install_procedure ("file_pac_save",
                          "Saves files in PAC file format",
                          "This plug-in saves PAC graphic format files.",
                          "Johan Bolmsjo <johan@nocrew.org>",
                          "Copyright 2002 by Johan Bolmsjo",
                          PLUG_IN_VERSION,
                          "<Save>/PAC",
                          "RGB*,GRAY*",
                          GIMP_PLUGIN,
                          nsave_args,
			  0,
                          save_args,
			  NULL);

  gimp_register_magic_load_handler ("file_pac_load",
				    "pac",
				    "",
				    "0,string,PAC0");
  
  gimp_register_save_handler       ("file_pac_save",
				    "pac",
				    "");
}


static void
run (const gchar   *name,
     gint     nparams,
     const GimpParam  *param,
     gint    *nreturn_vals,
     GimpParam **return_vals)
{
  static GimpParam values[2];
#ifdef PAC_GIMP_12
  GimpRunModeType  run_mode;
#else
  GimpRunMode run_mode;
#endif
  GimpPDBStatusType   status = GIMP_PDB_SUCCESS;
  gint32        image_ID;
  gint32        drawable_ID;
#ifdef PAC_GIMP_12
  GimpExportReturnType export = GIMP_EXPORT_CANCEL;
#else
  GimpExportReturn export = GIMP_EXPORT_CANCEL;
#endif
  int interactive = FALSE;
  
  run_mode = param[0].data.d_int32;

  *nreturn_vals = 1;
  *return_vals  = values;
  values[0].type          = GIMP_PDB_STATUS;
  values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;

  if (strcmp (name, "file_pac_load") == 0)
    {
#ifdef PAC_GIMP_12
       INIT_I18N_UI();
#endif

       switch (run_mode)
        {
        case GIMP_RUN_INTERACTIVE:
	  interactive = TRUE;
          break;

        case GIMP_RUN_NONINTERACTIVE:
          /*  Make sure all the arguments are there!  */
          interactive = FALSE;
	  if (nparams != 3)
            status = GIMP_PDB_CALLING_ERROR;
          break;

        default:
          break;
        }

       if (status == GIMP_PDB_SUCCESS)
	 {
	   image_ID = read_pac (param[1].data.d_string, interactive);

	   if (image_ID != -1)
	     {
	       *nreturn_vals = 2;
	       values[1].type         = GIMP_PDB_IMAGE;
	       values[1].data.d_image = image_ID;
	     }
	   else
	     status = GIMP_PDB_EXECUTION_ERROR;
	 }
    }
  else if (strcmp (name, "file_pac_save") == 0)
    {
#ifdef PAC_GIMP_12
      INIT_I18N_UI();
#endif

      image_ID    = param[1].data.d_int32;
      drawable_ID = param[2].data.d_int32;
      
      /* Export image if needed (change pixel layout). */
      switch (run_mode)
	{
	case GIMP_RUN_INTERACTIVE:
	case GIMP_RUN_WITH_LAST_VALS:
	  gimp_ui_init ("pac", FALSE);
	  export = gimp_export_image (&image_ID, &drawable_ID, "PAC",
				      (GIMP_EXPORT_CAN_HANDLE_RGB |
				       GIMP_EXPORT_CAN_HANDLE_GRAY |
				       GIMP_EXPORT_CAN_HANDLE_ALPHA));
	  if (export == GIMP_EXPORT_CANCEL)
	    {
	      values[0].data.d_status = GIMP_PDB_CANCEL;
	      return;
	    }
	  break;
	  
	default:
	  break;
	}

      switch (run_mode)
        {
        case GIMP_RUN_INTERACTIVE:
	  interactive = TRUE;
          break;

        case GIMP_RUN_NONINTERACTIVE:
          /*  Make sure all the arguments are there!  */
          interactive = FALSE;
	  if (nparams != 5)
            status = GIMP_PDB_CALLING_ERROR;
          break;

        case GIMP_RUN_WITH_LAST_VALS:
          interactive = FALSE;
          break;

        default:
          break;
        }
      
      if (status == GIMP_PDB_SUCCESS)
	status = write_pac (param[3].data.d_string, image_ID, drawable_ID,
			   interactive);
      
      if (export == GIMP_EXPORT_EXPORT)
	gimp_image_delete (image_ID);
    }
  else
    status = GIMP_PDB_CALLING_ERROR;
  
  values[0].data.d_status = status;
}


static gint32
read_pac (gchar *filename, int interactive)
{
  PacObject_t *pac;
  PacImage_t *image;
  GimpPixelRgn pixel_rgn;
  gint32 gimp_image;
  gint32 gimp_layer;
  int image_type;
  int layer_type;
  GimpDrawable *gimp_drawable;
  gchar *temp_buf;
  
  if (interactive)
    {
      temp_buf = g_strdup_printf (_("Loading %s:"), filename);
      gimp_progress_init (temp_buf);
      g_free (temp_buf);
    }
  
  /* Create PAC object */
  if (! (pac = pac_new_obj ()))
    {
      g_message (pac_get_error ());
      return -1;
    }
  
  /* Set callback handler for progress bar */
  pac_progress_callback (pac, (void (*)(gdouble))gimp_progress_update);
  
  /* Load PAC file */
  if (! (image = pac_decode_from_file (pac, filename)))
    {
      pac_delete_obj (pac);
      g_message (pac_get_error ());
      return -1;
    }
  
  /* GIMP stuff */
  switch (image->format)
    {
    case PAC_GRAY:
      image_type = GIMP_GRAY;
      layer_type = GIMP_GRAY_IMAGE;
      break;
      
    case PAC_GRAY_ALPHA:
      image_type = GIMP_GRAY;
      layer_type = GIMP_GRAYA_IMAGE;
      break;
      
    case PAC_RGB:
      image_type = GIMP_RGB;
      layer_type = GIMP_RGB_IMAGE;
      break;
      
    case PAC_RGB_ALPHA:
      image_type = GIMP_RGB;
      layer_type = GIMP_RGBA_IMAGE;
      break;
      
    default:
      g_message (_("PAC: Unknown colour format (%s)"), filename);
      return -1;
    }
  
  gimp_image = gimp_image_new (image->width, image->height, image_type);
  gimp_layer = gimp_layer_new (gimp_image, _("Background"),
			       image->width, image->height, layer_type,
			       100, GIMP_NORMAL_MODE);
  
  gimp_image_set_filename (gimp_image, filename);
  
  gimp_image_add_layer (gimp_image, gimp_layer, 0);
  gimp_drawable = gimp_drawable_get (gimp_layer);
  
  gimp_pixel_rgn_init (&pixel_rgn, gimp_drawable, 0, 0, gimp_drawable->width,
		       gimp_drawable->height, TRUE, FALSE);
  
  gimp_pixel_rgn_set_rect (&pixel_rgn, image->data, 0, 0,
			   gimp_drawable->width, gimp_drawable->height);
  
  gimp_drawable_flush (gimp_drawable);
  gimp_drawable_detach (gimp_drawable);
  
  pac_delete_obj (pac);
  pac_delete_image (image);
  
  return gimp_image;
}


static GimpPDBStatusType
write_pac (gchar  *filename,
	   gint32  gimp_image,
	   gint32  drawable_ID,
	   int interactive)
{
  PacObject_t *pac;
  PacImage_t *image;
  PacFormat_t format;
  GimpDrawable *gimp_drawable;
  GimpImageType image_type;
  GimpPixelRgn pixel_rgn;
  gchar *temp_buf;
  
  /* Check if we can save this image */
  gimp_drawable = gimp_drawable_get (drawable_ID);
  image_type = gimp_drawable_type (drawable_ID);
  gimp_pixel_rgn_init (&pixel_rgn, gimp_drawable, 0, 0, gimp_drawable->width,
		       gimp_drawable->height, FALSE, FALSE);
  
  switch (image_type)
    {
    case GIMP_GRAY_IMAGE :
      format = PAC_GRAY;
      break;
      
    case GIMP_GRAYA_IMAGE :
      format = PAC_GRAY_ALPHA;
      break;
      
    case GIMP_RGB_IMAGE :
      format = PAC_RGB;
      break;
      
    case GIMP_RGBA_IMAGE :
      format = PAC_RGB_ALPHA;
      break;
      
    default:
      g_message (_("PAC: Unsupported image type (%s)"), filename);
      return GIMP_PDB_EXECUTION_ERROR;
    }
  
  if (interactive)
    {
      temp_buf = g_strdup_printf (_("Saving %s:"), filename);
      gimp_progress_init (temp_buf);
      g_free (temp_buf);
    }
  
  /* Create PAC object */
  if (! (pac = pac_new_obj ()))
    {
      g_message (pac_get_error ());
      return -1;
    }
  
  /* Set callback handler for progress bar */
  pac_progress_callback (pac, (void (*)(gdouble))gimp_progress_update);
  
  /* Create PAC image */
  image = pac_new_image (format, 8, gimp_drawable->width,
			 gimp_drawable->height);
  
  if (! image)
    {
      g_message (pac_get_error ());
      return -1;
    }
  
  /* Copy image */
  gimp_pixel_rgn_get_rect (&pixel_rgn, image->data, 0, 0, gimp_drawable->width,
			   gimp_drawable->height);
  
  if (pac_encode_to_file (pac, image, filename))
    {
      pac_delete_obj (pac);
      pac_delete_image (image);
      g_message (pac_get_error ());
      return -1;
    }
  
  gimp_drawable_detach (gimp_drawable);
  
  pac_delete_obj (pac);
  pac_delete_image (image);
  
  return GIMP_PDB_SUCCESS;
}
