/**
 * Copyright (c) 2000 by Johan Bolmsjo.
 * Copyright (C) 1999 Andreas Umbach <marvin@dataway.ch>
 * 
 * @author Andreas Umbach (PNG loading code).
 * @author Johan Bolmsjo <johan@nocrew.org>
 */

#include <stdio.h>
#include <stdlib.h>
#include <png.h>
#include "pac.h"

/* PNG read function file descriptor */
static FILE *png_stream;

/**
 * PNG library read function. This is not multi-thread safe!
 *
 * @param png_ptr  ???
 * @param data     Where to put data
 * @param length   Number of bytes to read
 */
static void
png_read_func (png_structp png_ptr, png_bytep data, png_size_t length)
{
  fread (data, 1, length, png_stream);
}

/**
 * Returns 0 on success and -1 on failure.
 */

PacImage_t *
png_decode_from_file (const char *src)
{
  PacImage_t *image;
  PacFormat_t format;
  int bits_per_component;
  int bytes_per_pixel, bytes_per_line, png_colour_type;
  png_uint_32 width, height, y;
  png_structp png_ptr = 0;
  png_infop info_ptr = 0;
  png_byte *p;
  
  if ((png_stream = fopen (src, "rb")) == NULL)
    {
      pac_set_error ("(png) Failed to open %s", src);
      goto png_error;
    }
  
  png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (! png_ptr)
    {
      pac_set_error ("(png) Failed to create read struct");
      goto png_error;
    }
  
  info_ptr = png_create_info_struct (png_ptr);
  if (! info_ptr)
    {
      pac_set_error ("(png) Failed to create info struct");
      goto png_error;
    }
  
  png_set_read_fn (png_ptr, 0, png_read_func);

  png_read_info (png_ptr, info_ptr);
  png_get_IHDR (png_ptr, info_ptr, &width, &height,
		&bits_per_component, &png_colour_type, 0, 0, 0);
  
  if (bits_per_component != 8)
    {
      pac_set_error ("(png) Wrong bitdepth: %d", bits_per_component);
      goto png_error;
    }
  
  switch (png_colour_type)
    {
    case PNG_COLOR_TYPE_GRAY:
      format = PAC_GRAY;
      bytes_per_pixel = 1;
      break;
      
    case PNG_COLOR_TYPE_GRAY_ALPHA:
      format = PAC_GRAY_ALPHA;
      bytes_per_pixel = 2;
      break;
      
    case PNG_COLOR_TYPE_RGB:
      format = PAC_RGB;
      bytes_per_pixel = 3;
      break;
      
    case PNG_COLOR_TYPE_RGB_ALPHA:
      format = PAC_RGB_ALPHA;
      bytes_per_pixel = 4;
      break;
      
    default:
      pac_set_error ("(png) Wrong colour type!");
      goto png_error;
    }
  
  bytes_per_line = width * bytes_per_pixel;
  
  image = pac_new_image (format, bits_per_component, width, height);
  if (! image)
    goto png_error;
  
  p = image->data;
  for (y = 0; y < height; y++)
    {
      png_read_row (png_ptr, p, 0);
      p += bytes_per_line;
    }
  
  fclose (png_stream);
  png_destroy_read_struct (&png_ptr, &info_ptr, 0);
  
  return image;
  
  /* An error occurred! */
 png_error:
  
  if (png_stream)
    fclose (png_stream);
  
  if (png_ptr)
    png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp)NULL);
  
  return NULL;
}


int
png_encode_to_file (PacImage_t *image, const char *dst)
{
  FILE *file = NULL;
  png_infop info = NULL;
  png_structp png = NULL;
  png_bytep data;
  png_bytepp row_pointers = NULL;
  int i, png_colour_type;
  int colour_table[4] = { PNG_COLOR_TYPE_GRAY, PNG_COLOR_TYPE_GRAY_ALPHA,
			  PNG_COLOR_TYPE_RGB, PNG_COLOR_TYPE_RGB_ALPHA };
  
  if (image->bits_per_component != 8)
    {
      pac_set_error ("(png) Unsuported bitdepth: %d",
		     image->bits_per_component);
      goto png_error;
    }
  
  if ((file = fopen (dst, "wb")) == NULL)
    {
      pac_set_error ("(png) Failed to open %s", dst);
      goto png_error;
    }
  
  row_pointers = malloc (image->height * sizeof (png_bytep));
  if (! row_pointers)
    {
      pac_set_error ("(png) Out of memory!");
      goto png_error;
    }
  
  data = image->data;
  for (i = 0; i < image->height; i++)
    {
      row_pointers[i] = data;
      data += image->bytes_per_row;
    }

  if ((png = png_create_write_struct (PNG_LIBPNG_VER_STRING,
				      NULL, NULL, NULL)) == NULL)
   {
     pac_set_error ("(png) Could not create write structure");
     goto png_error;
   }
  
  if ((info = png_create_info_struct (png)) == NULL)
   {
     pac_set_error ("(png) Could not create info structure");
     goto png_error;
   }
  
  if (setjmp (png_jmpbuf (png)))
    {
      pac_set_error ("(png) Could set jump value for writing");
      goto png_error;
    }
  
  png_init_io (png, file);
  png_colour_type = colour_table[image->format];
  
  png_set_IHDR (png, info, image->width, image->height,
		image->bits_per_component, png_colour_type,
		PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
		PNG_FILTER_TYPE_DEFAULT);
  
  png_write_info (png, info);
  png_write_image (png, row_pointers);
  png_write_end (png, info);
  png_destroy_write_struct (&png, &info);
  
  free (row_pointers);
  fclose (file);
  
  return 0;
  
 png_error:
  
  if (file)
    fclose (file);
  
  if (row_pointers)
    free (row_pointers);
  
  png_destroy_write_struct (&png, &info);
  
  return -1;
}
